@@ -27,6 +27,7 @@ import attr
 
 from hypothesis import Phase, Verbosity, HealthCheck
 from hypothesis import settings as Settings
+from hypothesis._settings import local_settings
 from hypothesis.reporting import debug_report
 from hypothesis.internal.compat import Counter, ceil, hbytes, hrange, \
     int_to_text, int_to_bytes, benchmark_time, int_from_bytes, \
@@ -43,6 +44,7 @@ __tracebackhide__ = True
 
 
 HUNG_TEST_TIME_LIMIT = 5 * 60
+MAX_SHRINKS = 500
 
 
 @attr.s
@@ -208,7 +210,7 @@ class ConjectureRunner(object):
 
         for u, v in data.blocks:
             # This can happen if we hit a dead node when walking the buffer.
-            # In that case we alrady have this section of the tree mapped.
+            # In that case we already have this section of the tree mapped.
             if u >= len(indices):
                 break
             self.block_sizes[indices[u]] = v - u
@@ -248,7 +250,7 @@ class ConjectureRunner(object):
                 self.interesting_examples[key] = data
                 self.shrunk_examples.discard(key)
 
-            if self.shrinks >= self.settings.max_shrinks:
+            if self.shrinks >= MAX_SHRINKS:
                 self.exit_with(ExitReason.max_shrinks)
         if (
             self.settings.timeout > 0 and
@@ -405,7 +407,7 @@ class ConjectureRunner(object):
             self.event_call_counts[event] += 1
 
     def debug(self, message):
-        with self.settings:
+        with local_settings(self.settings):
             debug_report(message)
 
     def debug_data(self, data):
@@ -433,7 +435,7 @@ class ConjectureRunner(object):
         ))
 
     def run(self):
-        with self.settings:
+        with local_settings(self.settings):
             try:
                 self._run()
             except RunIsComplete:
@@ -661,6 +663,17 @@ class ConjectureRunner(object):
                 HealthCheck.large_base_example
             )
 
+        # If the language starts with writes of length >= cap then there is
+        # only one string in it: Everything after cap is forced to be zero (or
+        # to be whatever value is written there). That means that once we've
+        # tried the zero value, there's nothing left for us to do, so we
+        # exit early here.
+        for i in hrange(self.cap):
+            if i not in zero_data.forced_indices:
+                break
+        else:
+            self.exit_with(ExitReason.finished)
+
         self.health_check_state = HealthCheckState()
 
         count = 0
@@ -1018,7 +1031,7 @@ class Negated(object):
         self.tag = tag
 
 
-NEGATED_CACHE = {}
+NEGATED_CACHE = {}  # type: dict
 
 
 def negated(tag):
@@ -1470,6 +1483,21 @@ class Shrinker(object):
             t.buffer[:t.blocks[i][0]] in self.__shrinking_prefixes
         )
 
+    def is_payload_block(self, i):
+        """A block is payload if it is entirely non-structural: We can tinker
+        with its value freely and this will not affect the shape of the input
+        language.
+
+        This is mostly a useful concept when we're doing lexicographic
+        minimimization on multiple blocks at once - by restricting ourself to
+        payload blocks, we expect the shape of the language to not change
+        under us (but must still guard against it doing so).
+        """
+        return not (
+            self.is_shrinking_block(i) or
+            i in self.shrink_target.forced_blocks
+        )
+
     def lower_common_block_offset(self):
         """Sometimes we find ourselves in a situation where changes to one part
         of the byte stream unlock changes to other parts. Sometimes this is
@@ -1558,7 +1586,10 @@ class Shrinker(object):
         def reoffset_pair(pair, o):
             n = len(self.blocks)
             # Number of blocks may have changed, need to validate
-            valid_pair = [p for p in pair if p < n and int_from_block(p) > 0]
+            valid_pair = [
+                p for p in pair if p < n and int_from_block(p) > 0 and
+                self.is_payload_block(p)
+            ]
 
             if len(valid_pair) < 2:
                 return
@@ -1575,12 +1606,12 @@ class Shrinker(object):
 
         i = 0
         while i < len(self.blocks):
-            if not self.is_shrinking_block(i) and int_from_block(i) > 0:
+            if self.is_payload_block(i) and int_from_block(i) > 0:
                 j = i + 1
                 while j < len(self.shrink_target.blocks):
                     block_val = int_from_block(j)
                     i_block_val = int_from_block(i)
-                    if not self.is_shrinking_block(j) \
+                    if self.is_payload_block(j) \
                        and block_val > 0 and i_block_val > 0:
                         offset = min(int_from_block(i),
                                      int_from_block(j))
@@ -1673,7 +1704,7 @@ class Shrinker(object):
             # it was caused the behaviour we're trying to shrink.
             # Everything non-structural, we redraw uniformly at random.
             for i, (u, v) in enumerate(self.blocks):
-                if not self.is_shrinking_block(i):
+                if self.is_payload_block(i):
                     attempt_buf[u:v] = uniform(self.__engine.random, v - u)
             attempt = self.cached_test_function(attempt_buf)
             if self.__predicate(attempt):
@@ -1686,7 +1717,7 @@ class Shrinker(object):
                     sort_key(prev.buffer)
                 ):
                     # We have successfully shrunk the example past where
-                    # we started from. Now we begin the whole processs
+                    # we started from. Now we begin the whole process
                     # again from the new, smaller, example.
                     count = 0
                 else:
