@@ -61,7 +61,6 @@ class Minimizer(object):
         buffer.
 
         Return True if it succeeds as such.
-
         """
         assert isinstance(buffer, hbytes)
         assert len(buffer) == self.size
@@ -149,9 +148,7 @@ class Minimizer(object):
 
         So that's what this method does. It's a cheat to give us good shrinking
         of floating at low cost in runtime and only moderate cost in elegance.
-
         """
-
         # If the block is of the wrong size then we're certainly not using the
         # float encoding.
         if self.size != 8:
@@ -214,6 +211,28 @@ class Minimizer(object):
             lambda c: c == self.current_int or self.incorporate_int(c)
         )
 
+    def sort(self):
+        return self.incorporate(hbytes(sorted(self.current)))
+
+    # Sometimes data can be sorted, except that some elements need to remain
+    # in constant locations to preserve invariants that the minimizer can't be
+    # aware of. Attempt to do as much sorting as possible.
+    # This does not try to bring elements 'across' stationary elements.
+    def partial_sort(self):
+        any_sorting_done = False
+        ps = list(self.current)
+        for i in hrange(self.size - 1):
+            j = i + 1
+            while j > 0 and ps[j - 1] > ps[j]:
+                prev_list = ps[:]
+                ps[j], ps[j - 1] = ps[j - 1], ps[j]
+                if self.incorporate(hbytes(ps)):
+                    any_sorting_done = True
+                else:  # Restore to the last usable found example
+                    ps = prev_list
+                j = j - 1
+        return any_sorting_done
+
     def run(self):
         if not any(self.current):
             return
@@ -271,11 +290,13 @@ class Minimizer(object):
             first = False
             change_counter = self.changes
 
+            self.sort()
             self.float_hack()
             self.shift()
             self.shrink_indices()
             self.rotate_suffixes()
             self.minimize_as_integer()
+            self.partial_sort()
 
 
 def minimize(initial, condition, random, full=True):
@@ -293,7 +314,6 @@ def binsearch(_lo, _hi):
 
     This function is used purely for its side effects and returns
     nothing.
-
     """
     def accept(f):
         lo = _lo
